
require "terra"
require "worldparams"

local _G = _G

function makegame(maingame, lev_sid, lev_wreck, lev_ote)
local world = setmetatable({}, {__index = _G})
world._G = world
setfenv(1, world)


if not lev_sid then lev_sid = 10 end
if not lev_wreck then lev_wreck = 10 end
if not lev_ote then lev_ote = 10 end

local lev_upg_power = 2.5 * math.pow(1.2, lev_sid)
local lev_upg_enemy = 0.6 / math.pow(1.2, lev_ote)
local lev_upg_speed = 0.1 + 0.02 * lev_wreck

local avglev = (lev_sid + lev_wreck + lev_ote) / 3

local health_sid = math.max(100 - (lev_sid - avglev) * 15, 20)
local health_wreck = math.max(100 - (lev_wreck - avglev) * 15, 20)
local health_ote = math.max(100 - (lev_ote - avglev) * 15, 20)

local health_max = math.max(health_sid, health_wreck, health_ote)
local health_min = math.min(health_sid, health_wreck, health_ote)

if health_min == health_sid and health_sid <= 70 then critical = "sid" end
if health_min == health_wreck and health_wreck <= 70 then critical = "wreck" end
if health_min == health_ote and health_ote <= 70 then critical = "ote" end

print("Started new level")
print("Levels (S W 1):", lev_sid, lev_wreck, lev_ote)
print("HP (S W 1):", health_sid, health_wreck, health_ote)
print("Gun power", lev_upg_power)
print("Enemy power", lev_upg_enemy)
print("Speed", lev_upg_speed)
print("Critical", critical)



local sequences = {
 dustplosion = {len = 7, sloth = 3, norepeat = true},
}
local gtex_reserve = {}
local function gtex(id)
  if not gtex_reserve[id] then gtex_reserve[id] = Texture(id) end
  return gtex_reserve[id]
end
local function std_seq_get(self)
  local itex
  if self.seq.direction == -1 then
    itex = self.seq.len - math.mod(math.floor(self.cpos / self.seq.sloth), self.seq.len)
  else
    itex = math.mod(math.floor(self.cpos / self.seq.sloth), self.seq.len) + 1
  end
  return gtex((self.seq.prefix or self.name) .. "_" .. itex)
end
local function ssg_argh(self)
  return math.mod(math.floor(self.cpos / self.seq.sloth), self.seq.len) + 1
end
local function std_seq_increment(self, x, y)
  local t = ssg_argh(self)
  self.cpos = self.cpos + 1
  local n = ssg_argh(self)
  
  if self.seq.steplen then
    t = math.floor(t / self.seq.steplen)
    n = math.floor(n / self.seq.steplen)
  end
  
  if t ~= n and self.seq.step and x and y then
    local v = proxivol(x, y)
    
    if v ~= 0 then
      PlaySound(self.seq.step, v)
    end
  end
  
  if self.cpos_done and self.cpos >= self.cpos_done then self.cpos = self.cpos_done - 1 end
end
local function std_done(self)
  return self.cpos_done == self.cpos + 1
end
local function NewSequence(id)
  assert(sequences[id], id)
  
  local rv = {seq = sequences[id], name = id}
  if rv.seq.norepeat then
    rv.cpos = 0
    rv.cpos_done = rv.seq.sloth * rv.seq.len
  else
    rv.cpos = math.random(rv.seq.len * rv.seq.sloth)
  end
  rv.get = std_seq_get
  rv.increment = std_seq_increment
  rv.done = std_done
  
  return rv
end



local terra = terra_init(mappieces, entities)
terra.load(io.snatch("lev.lev.lua"))

local start
local crystal
local gt
local grunts = {}

for _, v in pairs(terra.entities()) do
  if v.id == "start" then
    start = v
  elseif v.id == "crystal" then
    crystal = v
  elseif v.id == "gt" then
    gt = v
  elseif v.id == "grunt" then
    table.insert(grunts, v)
  end
end
assert(start and crystal and gt)
if maingame then start = crystal end


ShowMouseCursor(false)

local xcenter, ycenter = start.x, start.y

local draw_parent = UI_CreateParent(1024/32, 768/32)
local draw_axis = CreateFrame("Frame", draw_parent)

local selected = {}
local tickers = {}
local player = {}
local enemy = {}
local dead = {}

local zaplines = {}
local crys = {}

local person_attach
local main_nameframe
do
  local nameframe = CreateFrame("Button")
  main_nameframe = nameframe
  local name = CreateFrame("Text", nameframe)
  name:SetText("All Characters")
  name:SetColor(1, 1, 1)
  name:SetPoint("TOPLEFT", UIParent, "TOPLEFT", 10, 10)
  function nameframe:Click()
    selected = {}
    for k in pairs(player) do
      if k.name then selected[k] = true end
    end
  end
  
  person_attach = CreateFrame("Frame")
  person_attach:SetPoint("TOPLEFT", name, "BOTTOMLEFT", 30, 0)
  person_attach:SetHeight(0)
  function person_attach:Draw()
  end
end

local function sel(targ)
  local shift = IsKeyDownFrame("shift_left") or IsKeyDownFrame("shift_right")
  
  if shift then
    if selected[targ] then selected[targ] = nil else selected[targ] = true end
  else
    selected = {[targ] = true}
  end
end

local function gunfire(self, target, range, delay, damage, r, g, b)
  if self.cooldown then
    self.cooldown = self.cooldown - 1
    if self.cooldown <= 0 then
      self.cooldown = nil
    else
      return
    end
  end
  
  local rgd = range * range
  local closest
  for v in pairs(target) do
    local dx, dy = self.x - v.x, self.y - v.y
    local dt = dx * dx + dy * dy
    if dt > rgd then continue end
    if closest and v.health > closest.health then continue end
    closest = v
  end
  
  if closest then
    closest:GetShot(damage, self)
    self.cooldown = delay
    
    table.insert(zaplines, {self.x, self.y, closest.x + math.random() * 0.8 - 0.4, closest.y + math.random() * 0.8 - 0.4, r = r, g = g, b = b, thick = 1.0})
    
    self.ang = math.atan2(closest.y - self.y, closest.x - self.x) / 3.14159 / 2 * 360
  end
  
  return closest
end

function makecritter(in_pic, in_pic2, x, y, siz)
  if not siz then siz = 1 end
  local play = {x = x, y = y, healthmax = 100, health = 100, speed = 0.1}
  local pic
  local pic2
  if in_pic then pic = Texture(in_pic) end
  if in_pic2 then pic2 = Texture(in_pic2) end
  
  local main_img = CreateFrame("Button", draw_axis)
  main_img:SetLayer(math.random())
  function main_img:Draw()
    gl.PushMatrix()
    gl.Translate(self:GetPointOnAxis("x", 0.5), self:GetPointOnAxis("y", 0.5), 0)
    gl.Rotate((play.ang or -90) + 90, 0, 0, 1)
    
    if pic then glutil.RenderCenteredSprite(pic, 0, 0, siz, siz) end
    if pic2 then glutil.RenderCenteredSprite(pic2, 0, 0, siz, siz) end
    if play.tex and not play.tex:done() then glutil.RenderCenteredSprite(play.tex:get(), 0, 0, siz, siz) end
    
    if selected[play] then
      glutil.RenderCenteredEmptyBox(0.5, 0.5, 1.0, 0, 0, 1, 1)
    end
    
    gl.PopMatrix()
  end
  main_img:SetWidth(siz)
  main_img:SetHeight(siz)
  
  local healthbar = CreateFrame("Frame", main_img)
  play.hbar = healthbar
  healthbar:SetPoint("BOTTOMLEFT", main_img, "TOPLEFT", 0.1, -0.1)
  healthbar:SetPoint("RIGHT", main_img, "RIGHT", -0.1, nil)
  healthbar:SetHeight(0.2)
  healthbar:SetBackgroundColor(0.5, 0.5, 0.5)
  
  local healthmini = CreateFrame("Frame", healthbar)
  healthmini:SetPoint("TOPLEFT", healthbar, "TOPLEFT", 0.03, 0.03)
  healthmini:SetPoint("BOTTOMRIGHT", healthbar, "BOTTOMRIGHT", -0.03, -0.03)
  healthmini:SetBackgroundColor(0, 0, 0)
  
  local healthindicator = CreateFrame("Frame", healthmini)
  healthindicator:SetPoint("TOPLEFT", healthmini, "TOPLEFT")
  healthindicator:SetPoint("BOTTOM", healthmini, "BOTTOM")
  healthindicator:SetBackgroundColor(1, 1, 1)
  
  function play:Tick()
    if self.health <= 0 then
      dead[self] = true
      assert(self.Die)
    end
    
    main_img:SetPoint("CENTER", draw_axis, "TOPLEFT", self.x, self.y)
    healthindicator:SetPoint("RIGHT", healthmini, self.health / self.healthmax, nil)
    if self.npfg then self.npfg:SetPoint("RIGHT", self.npbg, self.health / self.healthmax, nil) end
    
    if self.x_dest then
      local dx, dy = self.x_dest - self.x, self.y_dest - self.y
      local len = math.sqrt(dx * dx + dy * dy)
      
      self.ang = math.atan2(dy, dx) / 3.14159 / 2 * 360
      
      local nx, ny
      
      if len <= self.speed then
        nx, ny = self.x_dest, self.y_dest
        self.x_dest, self.y_dest = nil, nil
      else
        dx, dy = dx / len, dy / len
        dx, dy = dx * self.speed, dy * self.speed
        nx, ny = self.x + dx, self.y + dy
      end
      
      if terra.layers[1][math.floor(nx)][math.floor(ny)] then
        self.x, self.y = nx, ny
      else
        self.x_dest, self.y_dest = nil, nil
      end
    end
  end
  play:Tick()
  function play:DirectTo(x, y)
    self.x_dest = x
    self.y_dest = y
  end
  function play:Die()
    main_img:Hide()
    main_img:Detach() -- what you need in order to destroy
  end
  function play:GetShot(damage, from)
    if self.Alert then self:Alert(from) end
    self.health = self.health - damage
  end
  
  return play, main_img
end

function makeplayer(name, pic, x, y, healthmax, lev, warn)
  local play, main_img = makecritter("body_hero", "hair_" .. pic, x, y)
  play.speed = lev_upg_speed
  play.name = name
  
  play.healthmax, play.health = health_max, healthmax -- oh shut up
  
  local pusher = CreateFrame("Button")
  
  local nameplate = CreateFrame("Text", pusher)
  nameplate:SetText(name)
  nameplate:SetColor(1, 1, 1)
  nameplate:SetPoint("TOPLEFT", person_attach, "BOTTOMLEFT", 0, 5)
  
  person_attach = nameplate
  
  local npbg = CreateFrame("Frame", pusher)
  npbg:SetBackgroundColor(0.4, 0, 0)
  npbg:SetPoint("TOPLEFT", nameplate, "TOPLEFT", -5, 6)
  npbg:SetPoint("BOTTOM", nameplate, "BOTTOM", nil, 0)
  npbg:SetWidth(200)
  npbg:SetLayer(-1)
  
  if maingame == "full" then
    local tlev = CreateFrame("Text", pusher)
    tlev:SetPoint("RIGHT", npbg, "RIGHT", -10)
    tlev:SetPoint("BOTTOM", nameplate, "BOTTOM", 0)
    tlev:SetText(string.format("(level %d)", tostring(math.floor(lev))))
    tlev:SetSize(15)
    tlev:SetColor(1, 1, 1)
  end
  
  local npfg = CreateFrame("Frame", npbg)
  npfg:SetPoint("TOPLEFT", npbg, "TOPLEFT")
  npfg:SetPoint("BOTTOMRIGHT", npbg, "BOTTOMRIGHT")
  npfg:SetBackgroundColor(0.3, 0.3, 0.3)
  
  play.npfg, play.npbg = npfg, npbg
  
  local wframefunc
  if warn then
    local wframe = CreateFrame("Text") -- we actually anchor this to nothing so it's not active
    wframe:SetPoint("BOTTOMLEFT", npbg, "BOTTOMRIGHT", 10, 0)
    wframe:SetColor(1, 0.2, 0.2)
    
    local wff1 = coroutine.wrap(function ()
      while true do
        --wframe:SetText("KEEP SAFE")
        coroutine.pause(60)
        --wframe:SetText("DAMAGED")
        coroutine.pause(60)
      end
    end)
    local wff2 = coroutine.wrap(function ()
      local x = 0
      while true do
        x = x + 0.05
        npbg:SetBackgroundColor(0.6 + math.sin(x) * 0.15, 0, 0)
        coroutine.yield()
      end
    end)
    
    wframefunc = function ()
      wff1()
      wff2()
    end
    
    wframefunc()
  end
  
  
  local function selec()
    sel(play)
  end
  pusher.Click = selec
  main_img.Click = selec
  
  
  local otick = play.Tick
  function play:Tick()
    if wframefunc then wframefunc() end
    local shot = gunfire(self, enemy, 7, 3, lev_upg_power, 1, 0, 0)
    if shot then PlaySound("enemyshoot") end
    
    otick(self)
  end
  
  tickers[play] = true
  player[play] = true
  selected[play] = true
  
  return play
end

local kills = 0
function makeenemy(x, y)
  local play, main_img = makecritter("body_" .. math.random(7), "hair_" .. math.random(5), x, y)
  
  tickers[play] = true
  enemy[play] = true
  
  local otick = play.Tick
  function play:Tick()
    local targ = gunfire(self, player, 8, 10, lev_upg_enemy, 0, 1, 0)
    if targ then
      self:Alert(targ)
    end
    
    otick(self)
  end
  
  local odie = play.Die
  function play:Die()
    kills = kills + 1
    odie(self)
    
    PlaySound("boom")
    
    local boof = makecritter(nil, nil, self.x, self.y)
    boof.health = 0
    boof.tex = NewSequence("dustplosion")
    tickers[boof] = true
    
    function boof:Tick()
      self.tex:increment()
      if self.tex:done() then 
        dead[self] = true
      end
    end
    
    boof.hbar:Hide()
    
    if maingame then
      local cry = makecritter("gem", nil, self.x, self.y)
      cry.health = 1
      tickers[cry] = true
      crys[cry] = true
      cry.hbar:Hide()
    end
  end
  
  function play:Alert(from)
    if self.alerted then return end
    self.alerted = true
    
    local dx, dy = from.x - self.x, from.y - self.y
    local td = math.sqrt(dx * dx + dy * dy)
    dx, dy = dx / td, dy / td
    
    td = math.max(0, td - 4)
    
    dx, dy = dx * td, dy * td
    self:DirectTo(self.x + dx, self.y + dy)
    
    local thresh = 25
    for k in pairs(enemy) do
      if not k.Alert then continue end
      local tdx, tdy = k.x - self.x, k.y - self.y
      if tdx * tdx + tdy * tdy <= thresh then k:Alert(from) end
    end
  end
  
  return play
end

local sid = makeplayer("Sid", "sid", start.x + 0, start.y + -1, health_sid, lev_sid, critical == "sid") -- shot damage
local wreck = makeplayer("Wreck", "wreck", start.x + -1, start.y + 0, health_wreck, lev_wreck, critical == "wreck") -- speed
local ote = makeplayer("128", "128", start.x + 1, start.y + 0, health_ote, lev_ote, critical == "ote") -- enemy debuff auras

local crystal = makecritter("crystal", nil, crystal.x, crystal.y, 2)
tickers[crystal] = true

if not maingame then
  enemy[crystal] = true

  function crystal:GetShot()
    completion = {"crystal_captured"}
  end
else
  player[crystal] = true
  
  function crystal:Die()
    completion = {"crystal_destroyed"}
  end
  
  local next_spawn = 3
  function spawn()
    local wid = math.min(next_spawn * 1.5, 12)
    
    local spx, spy = gt.x, gt.y + 6
    
    for i = 1, next_spawn do
      local tx, ty = spx + math.random() * wid - wid / 2, spy + math.random()
      makeenemy(tx, ty):DirectTo(tx, crystal.y)
    end
    
    next_spawn = next_spawn + 1
  end
end

local gt_damage = 100 -- nuuuuuked

if maingame then gt_damage = lev_upg_enemy * 8 end

local gtp
do
  gtp = makecritter("gt", nil, gt.x, gt.y, 6)
  gtp.healthmax = 10000
  gtp.health = gtp.healthmax
  
  gtp.sid_cd = {x = gt.x, y = gt.y}
  gtp.wreck_cd = {x = gt.x, y = gt.y}
  gtp.ote_cd = {x = gt.x, y = gt.y}
  
  local otick = gtp.Tick
  function gtp:Tick()
    otick(self)
    
    gunfire(self.sid_cd, {[sid] = true}, 8, 3, gt_damage, 1, 1, 0)
    gunfire(self.wreck_cd, {[wreck] = true}, 8, 3, gt_damage, 1, 1, 0)
    gunfire(self.ote_cd, {[ote] = true}, 8, 3, gt_damage, 1, 1, 0)
    
    self.ang = -90
  end
  function gtp:Die()
    PlaySound("boom")
    
    self.hbar:Hide()
    
    for i = 1, 40 do
      self:SpawnPuff()
    end
  end
  function gtp:SpawnPuff()
    local boof = makecritter(nil, nil, self.x + (math.random() - 0.5) * 6, self.y + (math.random() - 0.5) * 6)
    boof.health = 0
    boof.tex = NewSequence("dustplosion")
    tickers[boof] = true
    
    function boof:Tick()
      self.tex:increment()
      if self.tex:done() then 
        dead[self] = true
      end
    end
    
    boof.hbar:Hide()
  end
  
  tickers[gtp] = true
  enemy[gtp] = true
end


if not maingame then
  for _, v in ipairs(grunts) do
    makeenemy(v.x, v.y)
  end
end


local function absorb()
  if maingame then
    for i = 1, 1000 do
      for k, _ in pairs(crys) do
        k.speed = i * 0.1
        k:DirectTo(eventual_completion.dt.x, eventual_completion.dt.y)
      end
      
      for k in pairs(tickers) do
        if crys[k] or k.tex then
          k:Tick()
        end
      end
      
      coroutine.yield()
      
      if i == 20 then completion = eventual_completion end
    end
  else
    completion = eventual_completion
  end
end

local function updatezaplines()
  local nz = {}
  for _, v in pairs(zaplines) do
    v.thick = v.thick - 0.1
    if v.thick > 0 then
      table.insert(nz, v)
    end
  end
  zaplines = nz
end

local function slain()
  for i = 1, 20 do
  
    updatezaplines()
    
    for k in pairs(tickers) do
      if k.tex then
        k:Tick()
      end
    end
    
    local ct = 12
    if i < 5 then ct = ct * 2 end
    
    for k = 1, ct do
      gtp:SpawnPuff()
    end
    
    coroutine.yield()
  end
  
  completion = eventual_completion 
end

local skrip

function tick_loop()
  if skrip then skrip() return end
  
  local speed = 0.5
  if IsKeyDownFrame("arrow_left") or IsKeyDownFrame("a") then xcenter = xcenter - speed end
  if IsKeyDownFrame("arrow_right") or IsKeyDownFrame("d") then xcenter = xcenter + speed end
  if IsKeyDownFrame("arrow_up") or IsKeyDownFrame("w") then ycenter = ycenter - speed end
  if IsKeyDownFrame("arrow_down") or IsKeyDownFrame("s") then ycenter = ycenter + speed end

  for k in pairs(dead) do
    if player[k] and k ~= crystal then
      eventual_completion = {"player_dead", k.name}
      if eventual_completion[2] == "128" then eventual_completion[2] = "ote" end
      if eventual_completion[2] == "Sid" then eventual_completion[2] = "sid" end
      if eventual_completion[2] == "Wreck" then eventual_completion[2] = "wreck" end
      eventual_completion.kills = kills
      eventual_completion.dt = k
      
      if not skrip then skrip = coroutine.wrap(absorb) end
    elseif k == gtp then
      eventual_completion = {"boss_slain"}
      if not skrip then skrip = coroutine.wrap(slain) end
      k:Die()
    else
      tickers[k] = nil
      player[k] = nil
      enemy[k] = nil
      k:Die()
    end
  end
  dead = {}
  
  updatezaplines()
  
  for k in pairs(tickers) do
    k:Tick()
  end
end
function render()
  glutil.SetScreen(xcenter - 1024/32/2, ycenter - 768/32/2, xcenter + 1024/32/2, ycenter + 768/32/2)
  
  terra.render_buckets(xcenter - 1024/32/2, ycenter - 768/32/2, xcenter + 1024/32/2, ycenter + 768/32/2)

  glutil.SetScreen(0, 0, 1024 / 32, 768 / 32)
  
  draw_axis:SetPoint("TOPLEFT", draw_parent, "CENTER", -xcenter, -ycenter)
  draw_parent:Render()
  --[[glutil.SetScreenCentered(xcenter, ycenter, 32)
  
  for k in pairs(tickers) do
    k:Render()
  end]]

  glutil.SetScreen(xcenter - 1024/32/2, ycenter - 768/32/2, xcenter + 1024/32/2, ycenter + 768/32/2)
  
  gl.LineWidth(3)
  gl.Enable("LINE_SMOOTH")
  
  gl.Begin("LINES")
  for _, v in pairs(zaplines) do
    gl.Color(v.r, v.g, v.b, v.thick)
    gl.Vertex(v[1], v[2])
    gl.Vertex(v[3], v[4])
  end
  gl.End()
  
  gl.LineWidth(1)
  gl.Disable("LINE_SMOOTH")
  
  glutil.ResetScreen()
  
  UIParent:Render()
end

local direct_it = {
  {{0, 0}},
  {{0.6, 0}, {-0.6, 0}},
  {{0, -1}, {1, 0}, {-1, 0}},
}
function key(button, ascii, event)
  if button == "mouse_right" and event == "press" then
    local x, y = GetMouseX() / 32 + xcenter - 1024/32/2, GetMouseY() / 32 + ycenter - 768/32/2
    local select_ct = 0
    for k in pairs(selected) do select_ct = select_ct + 1 end
    
    local select_it = 1
    for k in pairs(selected) do
      k:DirectTo(x + direct_it[select_ct][select_it][1], y + direct_it[select_ct][select_it][2])
      select_it = select_it + 1
    end
  end
  
  if button == "1" and event == "press" then sel(sid) end
  if button == "2" and event == "press" then sel(wreck) end
  if button == "3" and event == "press" then sel(ote) end
  if button == "`" and event == "press" then main_nameframe:Click() end
end

return world

end


local par = UI_CreateParent(1024, 768)
local diatex

loud = true

local dialogue = CreateFrame("Frame", par)
local brightness = CreateFrame("Frame", par)
brightness:SetAllPoints()
brightness:SetLayer(2)
brightness:Hide()
dialogue:Hide()

local resuf = CreateFrame("Text", par)
resuf:SetPoint("CENTER", par, "CENTER")
resuf:SetLayer(4)
resuf:Hide()

loud = false

local tbocks = CreateFrame("Frame", dialogue)
tbocks:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", 100, -100)
tbocks:SetPoint("RIGHT", UIParent, "RIGHT", -100)
tbocks:SetHeight(150)

local display = CreateFrame("Frame", tbocks)
display:SetPoint("TOPLEFT", tbocks, "TOPLEFT")
display:SetWidth(tbocks:GetHeight())
display:SetHeight(tbocks:GetHeight())
function display:Draw()
  if diatex then glutil.RenderBoundedSprite(diatex, self:GetBounds()) end
end

local werdbox = CreateFrame("Frame", tbocks)
werdbox:SetPoint("TOPLEFT", display, "TOPRIGHT", 30, 0)
werdbox:SetPoint("BOTTOMRIGHT", tbocks, "BOTTOMRIGHT")
werdbox:SetBackgroundColor(0, 0, 0)

local werds = CreateFrame("Text_Multiline", werdbox)
werds:SetPoint("TOPLEFT", werdbox, "TOPLEFT", 10, 10)
werds:SetPoint("BOTTOMRIGHT", werdbox, "BOTTOMRIGHT", -10, -10)


function ShowText(speaker, text)
  werds:SetText(text)
  diatex = Texture(speaker)
  
  coroutine.yield()
  while not WasKeyPressed("z") and not WasKeyPressed("mouse_left") and not WasKeyPressed(" ") and not WasKeyPressed("enter") and not WasKeyPressed("escape") do
    coroutine.yield()
  end
  
  PlaySound("chat")
end

function ShowConvo(list)
  dialogue:Show()
  for _, t in ipairs(list) do
    ShowText(t[1], t[2])
    if WasKeyPressed("escape") then break end
  end
  dialogue:Hide()
end

-- Conversations!
local conversations = {}
conversations.intro = {
  {"128", "Target \"GT\" isolated. Defensive matrix detected. Landing spot chosen. Approaching."},
  {"sid", "You think we've got them this time?"},
  {"wreck", "Maybe."},
  {"sid", "All these years. All that destruction. They need to be stopped."},
  {"wreck", "Yeah."},
  {"128", "Point of no return reached."},
  {"wreck", "No turning back."},
  {"128", "Unknown power signature detected."},
  {"sid", "The hell? 128, active scan. See what it is."},
  {"128", "Wild protosentient chronoenergy detected. No containment field."},
  {"sid", "Some kind of an experiment? What're they doing in there?"},
  {"wreck", "Doesn't matter. We gotta stop them."},
  {"128", "Approach complete."},
  {"sid", "Yeah. Let's go."},
  {"128", "Basic controls: Use WASD to look around the map. Right-click to issue commands. Left-click on your units to select them, or use the ~123 keys. Shift may be used to select multiple units."},
}



conversations.death = {}
conversations.death.ote = {
  {"128", "Armor compromised. Life support fai lin g. Sy st m   s t dn"},
  {"sid", "128! Dammit! We need him, or . . ."},
  {"wreck", "Too late. Missiles comin'."},
  {"sid", ". . . Damn. Should've kept him out of fire."},
}
conversations.death.sid = {
  {"sid", "Aaagh!"},
  {"128", "Scanning Sid. Armor compromise detected."},
  {"wreck", "You okay?"},
  {"sid", "Yeah . . . I think . . ."},
  {"128", "Reactor containment compromised."},
  {"sid", "Oh shit. The reactor."},
  {"wreck", "Shut it down."},
  {"128", "Reactor meltdown imminent."},
  {"sid", "I can't! It's going critical! He's taken too much damage!"},
  {"128", "Meltdown inevitable. Goodbye."},
}
conversations.death.wreck = {
  {"wreck", "Oof."},
  {"sid", "Everything alright?"},
  {"wreck", "No. They hit the computer."},
  {"128", "Motive field collapsing."},
  {"sid", "But that held back the stasis field. Without that, we're . . ."},
  {"wreck", "Stuck."},
  {"sid", "Forever."},
  {"128", "Computer damage excessive. Proper care was not taken. Motive f i e l d   c   o   l   l   a   p   s"},
}

conversations.victory_mid_1 = {
  {"sid", ". . . The hell is that?"},
  {"128", "Chronoenergy transfer detected. Chronoenergy is being extracted from GT."},
  {"sid", "GT! What was he using it for?"},
  {"128", "Unknown. Chronoenergy transfer nearly complete."},
  {"wreck", "Can't be bad for us."},
  {"128", "Chronoenergy transfer complete. Attempting to communicate with sentience."},
  {"wreck", "Well?"},
}

conversations.victory_mid_2 = {
  {"128", "GZZZZT"},
  {"sid", "AAAH"},
  {"wreck", "Urgh"},
  {"128", ". . . Systems online. Power drained."},
  {"sid", "What was that? I've never felt so . . ."},
  {"wreck", "Weak."},
  {"sid", "Yeah. So weak. Look, the rifle reactor's nearly offline."},
  {"wreck", "Motive field's almost gone too."},
  {"128", "Countermeasure module at seven percent. Power trapped in crystal by unknown force."},
  {"sid", "Maybe we can get our strength back. Let's blow up the crystal."},
  {"128", "Negative! Contained power is phenomenal. Uncontrolled release may prove catastrophic."},
  {"wreck", "We don't have time to work on it. Press on."},
  {"sid", "Alright. Let's move."},
}

conversations.minideath = {{}, {}, {}}
conversations.minideath[1].ote = {
  {"128", "Warning! All systems . . . nominal. Armor intact. Life support intact."},
  {"sid", "What? What are we doing here?"},
  {"128", "Unknown. We have been teleported to the crystal. Chronoenergy surge receding."},
  {"sid", "Are you okay?"},
  {"128", "Systems appear functional. Countermeasure module power increased."},
  {"wreck", "Increased!"},
  {"128", "Affirmative. Power has been returned from the crystal, using our deceased enemies as a catalyst. Enemy weaponry will have reduced effect."},
  {"sid", "Well, that's a good sign! It must have returned power to you after you . . . died."},
  {"128", "Life support . . . fluctuating. Power unstable. Integrity compromised."},
  {"wreck", "Good to fight?"},
  {"128", "Affirmative."},
  {"wreck", "Move on."},
}
conversations.minideath[1].sid = {
  {"sid", "Aaagh!"},
  {"wreck", "Sid!"},
  {"sid", "No, wait, it's . . . it's okay. Doesn't hurt. But what are we doing here?"},
  {"128", "Unknown. We have been teleported to the crystal. Chronoenergy surge receding. You appear undamaged."},
  {"sid", "Yeah. I seem fine. Hey, look at this! The reactor's powered up!"},
  {"128", "Weapon power increased. The crystal appears to have returned power to you, using our deceased enemies as a catalyst."},
  {"wreck", "That'll make things easier."},
  {"sid", "Oof. I still feel weak, though. Fragile."},
  {"128", "Armor structure appears compromised. Your resilience may be reduced."},
  {"sid", "Nothing we can do about it, though."},
  {"wreck", "Move on."},
}
conversations.minideath[1].wreck = {
  {"wreck", "Urph!"},
  {"128", "Lifesigns are . . . nominal. No damage detected."},
  {"sid", "Wreck! You okay?"},
  {"wreck", ". . . Yeah. Okay. Where are we?"},
  {"128", "Unknown. We have been teleported to the crystal. Chronoenergy surge receding."},
  {"wreck", "Motive field's powered up."},
  {"sid", "What? Why? Did the crystal give us our power back, and save us from . . . well, death?"},
  {"128", "Powerup occured at same time as chronoenergy surge. Power has been returned from the crystal, using our deceased enemies as a catalyst. Scanning armor. Armor structure appears compromised."},
  {"wreck", "I feel weaker."},
  {"sid", "Are you going to be okay?"},
  {"wreck", "I can fight. Move on."},
}

conversations.minideath[2].ote = {
  {"128", "We have been teleported again."},
  {"sid", "This is gonna happen a lot."},
  {"128", "Countermeasure module power increased. Enemy weaponry will have reduced effect."},
  {"wreck", "Good."},
  {"sid", "At this rate, we'll get all our power back soon."},
}
conversations.minideath[2].sid = {
  {"sid", "Here we are again."},
  {"128", "Affirmative. Chronoenergy surge receding."},
  {"sid", "Reactor power's back up."},
  {"wreck", "Good. Stronger guns will make this easier."},
  {"sid", "At this rate, we'll get all our power back soon."},
}
conversations.minideath[2].wreck = {
  {"sid", "Here we are again."},
  {"128", "Affirmative. Chronoenergy surge receding."},
  {"wreck", "Motive field's powered up."},
  {"sid", "At this rate, we'll get all our power back soon."},
}

conversations.minideath[3].ote = {
  {"128", "Chronoenergy surge receding. Countermeasure module power increased."},
  {"sid", "I feel stronger now."},
  {"wreck", "Same."},
  {"128", "My armor appears compromised."},
  {"sid", "But we're repaired. Interesting."},
}
conversations.minideath[3].sid = {
  {"sid", "Reactor power's moving up again."},
  {"128", "Chronoenergy surge receding. Armor structure reintegrating."},
  {"sid", "And now I'm feelin' weaker."},
  {"wreck", "Let's try again."},
}
conversations.minideath[3].wreck = {
  {"wreck", "Motive field increasing."},
  {"sid", "I feel stronger now."},
  {"128", "Chronoenergy surge receding. Armor structure reintegrating."},
  {"wreck", "And I'm weaker."},
  {"sid", "Interesting. Let's give this another shot."},
}


conversations.neardeath = {{}, {}}
conversations.neardeath[1].ote = {
  {"128", "Armor failure. Internal power source low. Energy drained."},
  {"sid", "128 can't take much more of that. If we're capable of dying, he nearly has."},
  {"wreck", "We're up front. You stay back. Far back. Don't get killed."},
}
conversations.neardeath[1].sid = {
  {"sid", "Ugh. So weak. I can barely stand."},
  {"128", "Detecting severe structural failure. Your armor has nearly shattered."},
  {"sid", "I can't take much more of this."},
  {"wreck", "We're up front. You stay back. Far back. Don't get killed."},
}
conversations.neardeath[1].wreck = {
  {"wreck", "Oof."},
  {"sid", "You okay?"},
  {"wreck", "No."},
  {"128", "Your motive field appears to be fluctuating badly, and is on the verge of destabilization. Caution is advised."},
  {"sid", "Stay well behind us, Wreck. We'll take care of it. Keep yourself alive."},
}

conversations.neardeath[2].ote = {
  {"128", "Armor nearly at failure. Retreating."},
}
conversations.neardeath[2].sid = {
  {"sid", "I'm hurtin' bad. You guys head up, I'll stay behind."},
}
conversations.neardeath[2].wreck = {
  {"wreck", "Armor's damaged."},
}

conversations.firstneardeath = {
  {"128", "Remember that you can select individual characters by left-clicking on them."},
}


conversations.crystaldeath = {
  {"128", "Chronoenergy fluctuation detected."},
  {"sid", "The . . . reactor's failing? It's shutting down entirely."},
  {"wreck", "The crystal."},
  {"sid", "They . . . must have gotten behind us. Can . . . barely stand."},
  {"128", "System failure. Core shutdown. Power off."},
}

conversations.victory = {
  {"128", "Lifesign terminated."},
  {"sid", "It can't be. GT's dead?"},
  {"128", "Scanning."},
  {"wreck", "Finally."},
  {"128", "Confirmed. GT is dead. We are victorious."},
}




local game
local result
local echo_keys = false
function recreate_game(...)
  UI_Reset()
  UI_ReregisterParent(par, 1024, 768)   -- ughhhh
  game = makegame(...)
end


local function continue_game(max_frame)
  echo_keys = true
  if not max_frame then max_frame = 4000000000 end -- uh-huh
  while not game.completion and max_frame >= 0 do
    game.tick_loop()
    coroutine.yield()
    max_frame = max_frame - 1
  end
  echo_keys = false
end
local flashlen = 10
local function flashup()
  brightness:Show()
  for i = 1, flashlen do
    brightness:SetBackgroundColor(1, 1, 1, i / flashlen)
    game.tick_loop()
    coroutine.yield()
  end
  brightness:Hide()
end
local function flashdown()
  brightness:Show()
  for i = 1, flashlen do
    brightness:SetBackgroundColor(1, 1, 1, 1 - i / flashlen)
    coroutine.yield()
  end
  brightness:Hide()
end
local function game_routine()
  recreate_game()
  
  ShowConvo(conversations.intro)
  
  continue_game()
  
  if game.completion[1] == "player_dead" then
    ShowConvo(conversations.death[game.completion[2]])
    return false
  end
  
  
  local comp = game.completion
  assert(comp[1] == "crystal_captured") 
  ShowConvo(conversations.victory_mid_1)
  PlaySound("flash")
  flashup()
  recreate_game(true, 0, 0, 0)
  flashdown()
  ShowConvo(conversations.victory_mid_2)
  
  local hasdied = {}
  local hdc = 0
  local hascritical = {}
  local firstneardeath = true
  
  local factor = 0.5
  local offset = 2
  
  local spxp = math.pow(offset, 1 / factor)
  local xp = {sid = spxp, ote = spxp, wreck = spxp}
  
  while true do
    
    local delay = 180
    
    -- main game loop, including spawning
    for i = 1, 4 do
      game.spawn()
      
      for k = 1, delay do
        game.tick_loop()
      end
    end
    
    while true do
      continue_game(delay)
      
      if game.completion then break end
      
      game.spawn()
      print("spawn shit")
      -- spawn shit
    end
    
    if game.completion[1] == "boss_slain" then
      break
    elseif game.completion[1] == "crystal_destroyed" then
      ShowConvo(conversations.crystaldeath)
      return false
    elseif game.completion[1] == "player_dead" then
    
      local comp = game.completion
      local crit = game.critical
      
      if crit == comp[2] then -- whooooops
        print("deathid", comp[2])
        ShowConvo(conversations.death[comp[2]])
        return false
      else
        local steal = 0.2
        -- first we average stuff out a bit
        
        xp[comp[2]] = xp[comp[2]] + comp.kills
        
        local totx = xp.sid + xp.wreck + xp.ote
        totx = totx * steal
        
        xp.sid = xp.sid * (1 - steal) + totx / 3
        xp.wreck = xp.wreck * (1 - steal) + totx / 3
        xp.ote = xp.ote * (1 - steal) + totx / 3
        
        PlaySound("flash")
        flashup()
        recreate_game("full", math.pow(xp.sid, factor) - offset, math.pow(xp.wreck, factor) - offset, math.pow(xp.ote, factor) - offset)
        game.tick_loop()
        flashdown()
        
        if not hasdied[comp[2]] then
          hdc = hdc + 1
          ShowConvo(conversations.minideath[hdc][comp[2]])
          hasdied[comp[2]] = true
        end
        
        print("crittic", game.critical)
        if game.critical then
          if not hascritical[game.critical] then
            ShowConvo(conversations.neardeath[1][game.critical])
            hascritical[game.critical] = true
            
            if firstneardeath then
              ShowConvo(conversations.firstneardeath)
              firstneardeath = false
            end
          else
            ShowConvo(conversations.neardeath[2][game.critical])
          end
        end
      end
    end
  end
  
  print("Boss slain, game won")
  ShowConvo(conversations.victory)
  
  return true
end
local function metagame_routine()
  while true do
    local rv = game_routine()
    
    brightness:Show()
    resuf:Show()
    brightness:SetBackgroundColor(0, 0, 0, 0.8)
    local resu = CreateFrame("Frame", resuf)
    
    local tx1 = CreateFrame("Text_Multiline", resu)
    tx1:SetPoint("CENTER", par, 0.5, 0.4)
    tx1:SetWidth(700)
    tx1:SetHeight(300)
    
    local tx2 = CreateFrame("Text", resu)
    tx2:SetPoint("CENTER", par, 0.5, 0.7)
    tx2:SetWidth(600)
    tx2:SetColor(1, 1, 1)
    tx2:SetSize(120)
    
    local function wigglything(text)
      PlaySound("objection")
      tx2:SetText(text)
      
      for i = 1, 20 do
        tx2:SetPoint("CENTER", par, 0.5, 0.7, (math.random() - 0.5) * (21 - i) * 3, (math.random() - 0.5) * (21 - i) * 3)
        coroutine.pause(4)
      end
    end
    
    if rv then
      -- display success
      
      coroutine.pause(30)
      
      tx1:SetText("     And thus, our heroes were able to repel GT from his throne and from the universe. Never again would such a scourge be known. The Defenders would be known far and wide, respected and worshipped, for their unbelievable and unlikely")
      
      coroutine.pause(120)
      
      wigglything("VICTORY")
      
      coroutine.pause(60)
    else
            -- display success
      
      coroutine.pause(30)
      
      tx1:SetText("     And thus, our heroes were defeated by GT, cast into oblivion and lost forever. Never again would a rebellion arise against the tyranny. The Defenders would be forgotten forever, thanks to their catastrophic")
      
      coroutine.pause(120)
      
      wigglything("FAILURE")
      
      coroutine.pause(60)
    end
    
    while not WasKeyPressed("z") and not WasKeyPressed("mouse_left") and not WasKeyPressed(" ") and not WasKeyPressed("enter") and not WasKeyPressed("escape") do
      coroutine.yield()
    end
    
    resu:Detach()
    resuf:Hide()
    brightness:Hide()
    
  end
end

local coro = coroutine.wrap(metagame_routine)
coro()

function tick_loop()
  coro()
end

local mousepointer = Texture("mptr")
function render()
  glutil.RenderBoundedSprite(Texture("starscape"), 0, 0, 1024, 768)
  
  game.render()
  
  par:Render()
  
  glutil.RenderCenteredSprite(mousepointer, GetMouseX(), GetMouseY(), 32, 32)
end

function key(...)
  if echo_keys then
    game.key(...)
  end
end
